// -*- C++ -*-

#pragma once

#include "lib/defs.h"

#include "gl_include_base.h"
#include "env.h"
#include "buffer.h"


/**
\addtogroup GL OpenGL interoperability
 \{
 */

namespace goopax
{
  namespace impl
  {

#ifdef __linux__
    typedef GLXContext (*GLXCREATECONTEXTATTRIBSARBPROC)(Display*, GLXFBConfig, GLXContext, bool, const Int*);
#endif


#ifdef __APPLE__
    struct glinitdata
    {
      CGLShareGroupObj sharegroup;
    };
#else
    struct glinitdata
    {
      GLXContext_type context;
      Display_type display;
    };
#endif


    
    namespace interface
    {
      extern "C"
      {
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_flush_gl2(goopax_device_impl* device) noexcept;
          GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_create_device_from_gl(goopax_device_impl**, const glinitdata*, envmode, goopax::impl::Uint, const char*, const goopax::impl::Uint8_t*) noexcept;
          GOOPAX_DLL_EXPORT(3.0.3) errno_type goopax_E_create_device_from_gl2(goopax_device_impl**, const glinitdata*, envmode, goopax::impl::Uint, void(*)(), const char*, const goopax::impl::Uint8_t*) noexcept;
      }
    }


    inline void check_gl_error()
    {
      int err = glGetError();
      if (err != GL_NO_ERROR)
	{
	  exwhat() << "Got OpenGL error " << err;
	  throw EX::gl();
	}
    }


    template <typename T, typename SIZE_TYPE>
    buffer<T, SIZE_TYPE> buffer<T, SIZE_TYPE>::create_from_gl(goopax_device device, GLuint GLres, BUFFER_FLAGS flags)
    {
        /// \ingroup GL
      /// Create goopax buffer from existing OpenGL buffer
      /// \param device goopax device
      /// \param GLres buffer ID of OpenGL buffer
      /// \param flags access type
      GLint glsize = -1;

      glBindBuffer(GL_ARRAY_BUFFER, GLres);
      glGetBufferParameteriv(GL_ARRAY_BUFFER, GL_BUFFER_SIZE, &glsize);
      check_gl_error();

      if (glsize % gettype<T>::size != 0)
	{
	  exwhat() << "Size of OpenGL buffer not divisible by element size " << gettype<T>::size;
	  throw EX::memory();
	}
      return buffer<T, SIZE_TYPE>(device, glsize/gettype<T>::size, flags, &GLres, INITTEX_FROMGL());
    }

    template <Uint DIM, typename T, Bool normalized>
    image_buffer<DIM, T, normalized> image_buffer<DIM, T, normalized>::create_from_gl(goopax_device device,
											    GLuint GLres,
											    IMAGE_FLAGS flags,
											    GLenum GLtarget,
											    GLint GLmiplevel)
    {
        /// \ingroup GL
      /// Creates image that is linked to existing OpenGL image.
      /// \param device goopax device that contains the OpenGL image
      /// \param GLres OpenGL resource ID
      /// \param flags access type. Must be one of IMAGE_READ_WRITE, IMAGE_READ_ONLY, IMAGE_WRITE_ONLY, IMAGE_WRITE_DISCARD
      /// \param GLtarget OpenGL target. Defaults to GL_TEXTURE_1D or GL_TEXTURE_2D for 1D/2D images.
      /// \param miplevel OpenGL miplevel of the texture. Defaults to 0.
        return image_buffer(INITTEX_IMPL(),
                            if_call(interface::goopax_E_image_create_gl2, device.get_impl(), nullptr, DIM, false, GLres, GLtarget, GLmiplevel, flags, image_buffer::channel_type_enum, image_buffer::num_channels, (Bool)normalized));
    }
      
    template <Uint DIM, typename T, Bool normalized>
    image_array_buffer<DIM, T, normalized> image_array_buffer<DIM, T, normalized>::create_from_gl(goopax_device device,
															  GLuint GLres,
															  IMAGE_FLAGS flags,
															  GLenum GLtarget,
															  GLint GLmiplevel)
    {
        /// \ingroup GL
      /// Creates image array that is linked to existing OpenGL image.
      /// \param device goopax device that contains the OpenGL image
      /// \param GLres OpenGL resource ID
      /// \param flags access type. Must be one of IMAGE_READ_WRITE, IMAGE_READ_ONLY, IMAGE_WRITE_ONLY, IMAGE_WRITE_DISCARD
      /// \param GLtarget OpenGL target. Defaults to GL_TEXTURE_1D or GL_TEXTURE_2D for 1D/2D images.
      /// \param miplevel OpenGL miplevel of the texture. Defaults to 0.
      return image_array_buffer(INITTEX_IMPL(),
                                if_call(interface::goopax_E_image_create_gl2, device, nullptr, DIM, true, GLres, GLtarget, GLmiplevel, flags, image_array_buffer::channel_type_enum, image_array_buffer::num_channels, (Bool)normalized));
    }
      
    inline void flush_gl_interop(goopax_device device)
    {
        /// \ingroup GL
      /// Must be called before calling OpenGL functions on shared OpenGL buffers or images that exist on the given device.
      interface::goopax_flush_gl2(device.get_impl());
    }

    inline goopax_device get_device_from_gl_common(envmode backend, const glinitdata glinit)
    {
      goopax_device ret;
      ret.impl = if_call(interface::goopax_E_create_device_from_gl2, &glinit, backend, Uint(GOOPAX_VERSION_ID), goopax_debugfunc, GOOPAX_CERT, sig);
      return ret;
    }
    
#ifdef _WIN32
    inline goopax_device get_device_from_gl(envmode backend, const HGLRC glCtx, const HDC hDC)
    {
        /// \ingroup GL
      /// Create device from OpenGL context
      return get_device_from_gl_common(backend, {glCtx, hDC});
    }

    inline goopax_device get_device_from_gl(envmode backend = env_GPU)
    {
        /// \ingroup GL
      /// Create device from OpenGL context
      return get_device_from_gl(backend, wglGetCurrentContext(), wglGetCurrentDC());
    }
#elif defined(__linux__)
    inline goopax_device get_device_from_gl(envmode backend, const GLXContext_type glCtx, const Display_type display)
    {
        /// \ingroup GL
      /// Create device from OpenGL context
      return get_device_from_gl_common(backend, {glCtx, display});
    }

    inline goopax_device get_device_from_gl(envmode backend = env_GPU)
    {
        /// \ingroup GL
      /// Create device from OpenGL context
      return get_device_from_gl(backend, glXGetCurrentContext(), glXGetCurrentDisplay());
    }
#elif defined(__APPLE__)
    inline goopax_device get_device_from_gl(envmode backend = env_GPU, const CGLShareGroupObj sharegroup = CGLGetShareGroup(CGLGetCurrentContext()))
    {
        /// \ingroup GL
      /// Create device from OpenGL context
      return get_device_from_gl_common(backend, {sharegroup});
    }
#endif
  }
  using impl::flush_gl_interop;
  using impl::get_device_from_gl;
}

/// \}
