// -*- C++ -*-

#pragma once

#include "lib/range.h"
#include "lib/vec.h"

#include "gpu_defs.h"
#include "lib/param.h"
#include "reinterpret.h"
#include "il_writer_utils.h"
#include "image.h"
#include "env.h"
#include "loops.h"

/**
 \addtogroup kernel
GPU kernels are inherited from kernel. The entry point to the GPU kernel is provided by the function 'program'. However, GPU kernels can call other functions, so they are not limited to the body of function 'program'.

 Example:
 \code{.cpp}
 struct my_kernel :
    kernel<my_kernel>
 {
    void program(resource<int>& A)
    {
        A[global_id()] = 5;   // This is executed on the GPU
    }
 };
 \endcode
\{
*/

namespace goopax
{
  namespace impl
  {
    struct tex_arg_base;
    struct kernel_impl;
    struct full_kernel;
    struct rundata;
    struct goopax_future_base;
      template <typename G> typename G::gpu_value_type work_group_reduce(typename G::gpu_value_type value, Tuint width);

      namespace ns_param_kernel
      {
          template <class R, class T> struct pas;
          template <class R, class T, class REST, bool is_gpu, bool is_resource, bool is_gather> struct pas_split;
      }
      
    static GOOPAX_CONSTEXPR unsigned int calc_lcd(const unsigned int a, const unsigned int b, const unsigned int loop=2)
    {
      return ((a%loop == 0) && (b%loop == 0)) ?
	loop*calc_lcd(a/loop, b/loop, loop)
	:
	((loop >= a || loop >= b) ?
	 a*b
	 :
	 calc_lcd(a, b, loop+1));
    }
	

    template <typename A, typename B> struct tuple_append;

    template <typename B, typename... A> struct tuple_append<tuple<A...>, B>
    {
      using type = tuple<A..., B>;
    };

    using std::tuple_size;
    using std::get;
    
    
    template <class T> struct unbool
    {
      using type = T;
    };
    template<> struct unbool<gpu_bool>
    {
      using type = gpu_int;
    };
    template<> struct unbool<Tbool>
    {
      using type = Tint;
    };



    namespace interface
    {
      extern "C"
      {
	GOOPAX_DLL_EXPORT(3.0.0) Size_t goopax_fk_current_global_size2(kernel_impl*, goopax_device_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) Size_t goopax_fk_current_local_size2(kernel_impl*, goopax_device_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) Size_t goopax_fk_current_num_subthreads2(kernel_impl*, goopax_device_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) Size_t goopax_fk_current_num_groups2(kernel_impl*, goopax_device_impl*) noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_fk_force_compile3(kernel_impl*, goopax_device_impl*) noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_fk_arun3(rundata**, goopax_device_impl*, kernel_impl*, texbuf* const*, unsigned int, const paramtype_t*, unsigned int) noexcept;
          GOOPAX_DLL_EXPORT(3.0.0) void goopax_fk_read_param(gpu_writer_base_impl*, unsigned int) noexcept;
          GOOPAX_DLL_EXPORT(3.0.0) void goopax_fk_write_gather(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_fk_wait2(kernel_impl*, goopax_device_impl* device) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_rd_release2(rundata* rd) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_future_get3(rundata* rd, Size_t*, const paramtype_t**) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_rd_finished2(bool*, rundata*) noexcept;
#if !GOOPAX_DISTRIB
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_fk_get_disassembly2(const char**, goopax_device_impl*, kernel_impl*) noexcept;
#endif
	GOOPAX_DLL_EXPORT(3.0.0) unsigned int goopax_fk_get_num_writeloops2(const kernel_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) unsigned int goopax_fk_get_current_writeloop2(const kernel_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_fk_require_writeloops2(kernel_impl*, unsigned int) noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) void goopax_fk_force_set_local_size2(kernel_impl*, goopax_device_impl*, Int) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_fk_force_set_num_groups2(kernel_impl*, goopax_device_impl*, Int) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_fk_force_set_compfac2(kernel_impl*, goopax_device_impl*, Int) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_fk_force_set_global_size2(kernel_impl*, goopax_device_impl*, Int) noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_fk_new2(kernel_impl**, full_kernel*, const char*, const Uint8_t*, void(*)(full_kernel*, Tuint*), const char*(*)(full_kernel*), unsigned int, unsigned int) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_fk_set_this2(kernel_impl*, full_kernel*) noexcept;
        }
    }

      struct pas_data
      {
	
	template <class V> static GOOPAX_CONSTEXPR unsigned int paramsize()
	{
	  return (calc_lcd(gettype<typename unbool<V>::type>::size, gettype<paramtype_t>::size)/gettype<paramtype_t>::size);
	}
	template <class V> static V read(const Tuint parampos)
	{
	  static const Size_t lcd = calc_lcd(gettype<typename unbool<V>::type>::size, gettype<paramtype_t>::size);
	  array<typename gettype<V>::template change<paramtype_t>::type, (lcd / gettype<paramtype_t>::size)> data;
	  for (Tuint k=0; k<data.size(); ++k)
	    {
	      interface::goopax_fk_read_param(data[k].impl.get(), parampos+k);
	    }
	  return read<V>(data, 0);
	}
	template <class V, typename POS, class PD> static V read(const PD& paramres, const POS parampos)
	{
	  static const Size_t lcd = calc_lcd(gettype<typename unbool<V>::type>::size, gettype<paramtype_t>::size);
	  array<typename gettype<V>::template change<paramtype_t>::type, (lcd / gettype<paramtype_t>::size)> data;
	  for (unsigned int k=0; k<data.size(); ++k)
	    {
	      data[k] = paramres[parampos+k];
	    }
	  auto rettab = reinterpret<array<typename unbool<V>::type, (lcd / gettype<typename unbool<V>::type>::size)> >(data);
	  return static_cast<V>(rettab[0]);
	}

          template <class T> static void write_gather(const gpu_uint parampos, const T& value)
          {
              static const Size_t lcd = calc_lcd(gettype<typename unbool<T>::type>::size, gettype<paramtype_t>::size);
              array<typename gettype<T>::template change<paramtype_t>::type, (lcd / gettype<paramtype_t>::size)> data;
              write(data, 0, value);
              for (Tuint k=0; k<data.size(); ++k)
              {
                  gpu_uint pos = parampos + k;
                  interface::goopax_fk_write_gather(pos.impl.get(), data[k].impl.get());
              }
          }

	template <class T, typename POS, class PD> static void write(PD& paramdata, const POS parampos, const T& value)
	{
	  using pt_t = typename gettype<T>::template change<paramtype_t>::type;
	  static const Size_t lcd = calc_lcd(gettype<typename unbool<T>::type>::size, gettype<paramtype_t>::size);
	  array<typename unbool<T>::type, (lcd / gettype<typename unbool<T>::type>::size)> valuetab;
	  valuetab.fill(value);
	  valuetab[0] = value;

	  auto data = reinterpret<array<pt_t, (lcd / gettype<pt_t>::size)> >(valuetab);
	  for (unsigned int k=0; k<data.size(); ++k)
	    {
	      paramdata[parampos+k] = data[k];
	    }
	}
      };



    template <class FOO=void> void fk_call_impl(full_kernel*, Tuint*);
    template <class FOO=void> const char* fk_get_name(full_kernel*);
    

    struct goopax_future_base
    {
      rundata* rd = nullptr;
      
      goopax_future_base() = default;
      goopax_future_base(const goopax_future_base&) = delete;
      goopax_future_base& operator = (const goopax_future_base&) = delete;
      inline goopax_future_base(goopax_future_base&& b);
      inline goopax_future_base& operator = (goopax_future_base&& b);
      inline ~goopax_future_base();
    };
    

    template <typename OP> typename OP::value_type collectfunc(const paramtype_t* p, Tsize_t threads)
    {
      auto ret = pas_data::read<typename OP::value_type>(p, 0);
      for (unsigned int k=1; k<threads; ++k)
	{
	  ret = OP::op(ret, pas_data::read<typename OP::value_type>(p, k*pas_data::paramsize<typename OP::value_type>()));
	}
      return ret;
    }


      /// \ingroup kernel
      /// future class for gather return values
      /// \tparam T return type. May be a C++ intrinsic type, or a goopax struct type
    template <typename T> class goopax_future :
      protected goopax_future_base
    {
      template <class R, class U> friend struct ns_param_kernel::pas;
      template <class R, class U, class REST, bool is_gpu, bool is_resource, bool is_gather> friend struct ns_param_kernel::pas_split;

        
        T data;
      Tuint offset;
      T(*collect)(const paramtype_t*, Tsize_t);
      
      void set(rundata* rd0, Tuint offset0, T(*collect0)(const paramtype_t*, Tsize_t))
      {
	this->rd = rd0;
	this->offset = offset0;
	this->collect = collect0;
      }

    public:

      bool finished() const
      {
          /// Test if value is available
	if (rd == nullptr)
	  return true;
	else
	  return if_call(interface::goopax_E_rd_finished2, rd);
      }
      T get()
      {
          /// wait for kernel to finish and get value
	if (rd != nullptr)
	  {
	    Size_t nt;
	    const paramtype_t* p;
	    check_errno(interface::goopax_E_future_get3(rd, &nt, &p));
	    data = collect(p + offset*nt, nt);
	    rd = nullptr;
	  }
	return data;
      }
        /// default constructor
      goopax_future() = default;
      
      goopax_future(const goopax_future&) = delete;
      goopax_future& operator = (const goopax_future&) = delete;

        /// move semantics
        goopax_future(goopax_future&&) = default;
        /// move semantics
      goopax_future& operator = (goopax_future&&) = default;
    };


    goopax_future_base::goopax_future_base(goopax_future_base&& b) :
      rd(b.rd)
    {
      b.rd = nullptr;
    }

    goopax_future_base& goopax_future_base::operator = (goopax_future_base&& b)
    {
      if (rd != nullptr)
	{
	  interface::goopax_rd_release2(rd);
	}
      rd = b.rd;
      b.rd = nullptr;
      return *this;
    }

    goopax_future_base::~goopax_future_base()
    {
      if (rd != nullptr)
	{
	  interface::goopax_rd_release2(rd);
	}
    }

    

      /// \ingroup kernel
      /// kernel base
    struct full_kernel
    {
      unique_ptr<kernel_impl, interface::goopax_deleter> impl_p;
      
        /// \name thread numbers of last kernel call
        /// These were the thread numbers when the kernel was last called.
        /// \{

        /// global size
      inline Size_t current_global_size(goopax_device device) { return interface::goopax_fk_current_global_size2(impl_p.get(), device.get_impl()); }
        /// local size
      inline Size_t current_local_size(goopax_device device) { return interface::goopax_fk_current_local_size2(impl_p.get(), device.get_impl()); }
        /// instruction level parallelization
      inline Size_t current_num_subthreads(goopax_device device) { return interface::goopax_fk_current_num_subthreads2(impl_p.get(), device.get_impl()); }
        /// number of groups
      inline Size_t current_num_groups(goopax_device device) { return interface::goopax_fk_current_num_groups2(impl_p.get(), device.get_impl()); }
        /// \}
        
        
        /// \name overriding default thread numbers
        ///\{

        /// forces the local size for this device to the given value
      inline void force_local_size(goopax_device device, Tint local_size) { interface::goopax_fk_force_set_local_size2(impl_p.get(), device.get_impl(), local_size); }
        /// forces the global size for this device to the given value
        inline void force_global_size(goopax_device device, Tint global_size) { interface::goopax_fk_force_set_global_size2(impl_p.get(), device.get_impl(), global_size); }
        /// forces the number of workgroups for this device to the given value
        inline void force_num_groups(goopax_device device, Tint num_groups) { interface::goopax_fk_force_set_num_groups2(impl_p.get(), device.get_impl(), num_groups); }
      inline void force_compfac(goopax_device device, Tint compfac) {	interface::goopax_fk_force_set_compfac2(impl_p.get(), device.get_impl(), compfac); }
        /// \}
        
      inline string name() const { return pretty_typename(typeid(*this));}
        /// force compilation of kernel. Only useful for performance measurements and other occasions.
        inline void force_compile(goopax_device device) { check_errno(interface::goopax_E_fk_force_compile3(impl_p.get(), device.get_impl())); }
      
      virtual void implcall(Tuint*) = 0;

      void wait(goopax_device device = nullptr)
      {
          /// Wait for all kernel invocations that are currently in the queue will finish.
          check_errno(interface::goopax_E_fk_wait2(impl_p.get(), device.get_impl()));
      }
      
#if !GOOPAX_DISTRIB
      inline string get_disassembly(goopax_device device) const { return if_call(interface::goopax_E_fk_get_disassembly2, device.get_impl(), impl_p.get()); }
#endif

        inline unsigned int num_writeloops() const { return interface::goopax_fk_get_num_writeloops2(impl_p.get()); }
      inline unsigned int writeloop() const { return interface::goopax_fk_get_current_writeloop2(impl_p.get()); }
      inline void require_writeloops(unsigned int num) const { interface::goopax_fk_require_writeloops2(impl_p.get(), num); }
      
      full_kernel() = default;
      
      inline full_kernel(full_kernel&& b) noexcept
      {
          /// move semantics
	*this = move(b);
      }
      inline full_kernel& operator = (full_kernel&& b) noexcept
      {
          /// move semantics
	impl_p = move(b.impl_p);
	interface::goopax_fk_set_this2(impl_p.get(), this);
	return *this;
      }

      full_kernel(const full_kernel&) = delete;
      full_kernel& operator = (const full_kernel&) = delete;
    };


    template <class FOO> void fk_call_impl(full_kernel* K, Tuint* p)
    {
      K->implcall(p);
    }
    template <class FOO> const char* fk_get_name(full_kernel* K)
    {
      static string ret;
      ret = K->name();
      return ret.c_str();
    }


    

    
    struct gather_basecheck
    {
    };

    // FIXME: return vectors instead.
    template <class T> struct gather_base :
      public gather_basecheck,
      gettype<T>::gpu
    {
      using gpu_value_type = typename gettype<T>::gpu;
      using value_type = T;
        static_assert(!is_same<gpu_value_type, value_type>::value, "gather_base: It seems that you have used a gpu type as template argument, please use a cpu type.");
      
      gather_base()
      {}

      gather_base(const gpu_value_type& val) :
	gpu_value_type(val)
      {}
    };
    

    namespace ns_param_kernel
    {
      template <class T> struct is_resource_type :
	public false_type
      {};
      
      template <class T, typename size_type, bool CONST> struct is_resource_type<resource<T, size_type, CONST> > :
	public true_type
      {};

      template <unsigned int DIM, typename T, bool normalized> struct is_resource_type<image_resource<DIM, T, normalized> > :
	public true_type
      {};
      template <unsigned int DIM, typename T, bool normalized> struct is_resource_type<image_array_resource<DIM, T, normalized> > :
	public true_type
      {};

      typedef goopax::impl::Tuint pas_base_type;

      template <class R, class gpu_T, class REST> struct pas_split_common
      {
	static_assert(is_same<gpu_T, typename gettype<gpu_T>::gpu>::value, "Bad kernel parameter type. Perhaps you used an intrinsic type where you should have a gpu type in the program(...) function of your kernel?");
	typedef typename unrangetype<typename gettype<gpu_T>::cpu>::type T;
	static GOOPAX_CONSTEXPR unsigned int param_size() { return pas_data::paramsize<gpu_T>() + pas<R, REST>::param_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_size() { return pas<R, REST>::gather_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_num() { return pas<R, REST>::gather_num(); }
	static const unsigned int num_tex_args = pas<R, REST>::num_tex_args;
	template <class K, class... P> static void impl(K& kernel, P&... p)
	{
	  const auto data = pas_data::read<gpu_T>(pas<R, REST>::param_size());
	  pas<R, REST>::impl(kernel, p..., data);
	}
	template <class K, class V, typename PARAM, class RP, class... P> static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, Size_t vecpos, const T& b, P... p)
	{
	  pas_data::write(param, pas<R, REST>::param_size(), b);
	  pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos, p...);
	}
	template <class RP, class... P> static inline void setcollect(rundata* rd, RP rp, const T&, P... p)
	{
	  pas<R, REST>::setcollect(rd, rp, p...);
	}
      };

      template <class R, class MEM, class REST> struct pas_split<R, MEM, REST, false, true, false>
      {
	static_assert(sizeof(MEM)==0, "Please pass resources as references or const references to the GPU kernel!\ne.g.: program(resource<int>& a);");
      };


      template <typename BUFREF, typename R, typename MEM, typename REST>
      struct pas_common_resource
      {
	static GOOPAX_CONSTEXPR unsigned int param_size() { return pas<R, REST>::param_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_size() { return pas<R, REST>::gather_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_num() { return pas<R, REST>::gather_num(); }
	static const unsigned int num_tex_args = pas<R, REST>::num_tex_args + 1;
	template <class K, class... P> static void impl(K& kernel, P&... p)
	{
	  MEM res;
	  pas<R, REST>::impl(kernel, p..., res);
	}
	template <class K, class V, typename PARAM, class RP, class... P> static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, Size_t vecpos, BUFREF b, P... p)
	{
	  vec[vecpos] = b.get_buf();
	  pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos+1, p...);
	}
	template <class RP, class... P> static inline void setcollect(rundata* rd, RP rp, BUFREF, P... p)
	{
	  pas<R, REST>::setcollect(rd, rp, p...);
	}
      };
      

      template <class R, typename MEM1, typename MEM2, class REST> struct pas_split<R, resource<MEM1, MEM2, false>&, REST, false, true, false> :
	public pas_common_resource<buffer<MEM1, MEM2>&, R, resource<MEM1, MEM2, false>, REST>
      {};
      
      template <class R, typename MEM1, typename MEM2, class REST, bool CONST> struct pas_split<R, const resource<MEM1, MEM2, CONST>&, REST, false, true, false> :
	public pas_common_resource<const buffer<MEM1, MEM2>&, R, resource<MEM1, MEM2, CONST>, REST>
      {};
      
      template <typename BUFREF, class R, typename IMAGE, class REST> struct pas_common_image
      {
	static GOOPAX_CONSTEXPR unsigned int param_size() { return pas<R, REST>::param_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_size() { return pas<R, REST>::gather_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_num() { return pas<R, REST>::gather_num(); }
	static const unsigned int num_tex_args = pas<R, REST>::num_tex_args + 1;
	template <class K, class... P> static void impl(K& kernel, P&... p)
	{
	  IMAGE res;
	  pas<R, REST>::impl(kernel, p..., res);
	}
	template <class K, class V, typename PARAM, class RP, class... P> static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, Size_t vecpos, BUFREF b, P... p)
	{
	  vec[vecpos] = b.get_buf();
	  pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos+1, p...);
	}
	template <class RP, class... P> static inline void setcollect(rundata* rd, RP rp, BUFREF, P... p)
	{
	  pas<R, REST>::setcollect(rd, rp, p...);
	}
      };

      template <class R, unsigned int DIM, typename T, bool normalized, class REST> struct pas_split<R, image_resource<DIM, T, normalized>&, REST, false, true, false> :
	public pas_common_image<image_buffer<DIM, T, normalized>&, R, image_resource<DIM, T, normalized>, REST>
      {};
      
      template <class R, unsigned int DIM, typename T, bool normalized, class REST> struct pas_split<R, const image_resource<DIM, T, normalized>&, REST, false, true, false> :
	public pas_common_image<const image_buffer<DIM, T, normalized>&, R, image_resource<DIM, T, normalized>, REST>
      {};
      
      template <class R, unsigned int DIM, typename T, bool normalized, class REST> struct pas_split<R, image_array_resource<DIM, T, normalized>&, REST, false, true, false> :
	public pas_common_image<image_array_buffer<DIM, T, normalized>&, R, image_array_resource<DIM, T, normalized>, REST>
      {};
      
      template <class R, unsigned int DIM, typename T, bool normalized, class REST> struct pas_split<R, const image_array_resource<DIM, T, normalized>&, REST, false, true, false> :
	public pas_common_image<const image_array_buffer<DIM, T, normalized>&, R, image_array_resource<DIM, T, normalized>, REST>
      {};
      
        
      template <class R, class gpu_T, class REST> struct pas_split<R, gpu_T, REST, true, false, false> :
	public pas_split_common<R, gpu_T, REST>
      {};

      template <class R, class gpu_T, class REST> struct pas_split<R, gpu_T, REST, false, false, false> :
	public pas_split_common<R, gpu_T, REST>
      {};


      template <class R, class G, class REST> struct pas_split<R, G&, REST, false, false, true>
      {
	using gather_type = typename G::value_type;
	static GOOPAX_CONSTEXPR unsigned int param_size() { return pas<R, REST>::param_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_size() { return pas<R, REST>::gather_size() + pas_data::paramsize<gather_type>(); }
	static GOOPAX_CONSTEXPR unsigned int gather_num() { return pas<R, REST>::gather_num() + 1; }
	static const unsigned int num_tex_args = pas<R, REST>::num_tex_args;

	template <class K, class... P> static void impl(K& kernel, P&... p)
	{
	  G val;
	  pas<R, REST>::impl(kernel, p..., val);
	  typename gettype<gather_type>::gpu tot = work_group_reduce<G>((typename gettype<gather_type>::gpu)val, local_size());
	  gpu_if(local_id() == 0)
	    pas_data::write_gather(pas<R, REST>::gather_size()*num_groups() + group_id()*pas_data::paramsize<gather_type>(), tot);
	}
	template <class K, class V, typename PARAM, class RP, class... P> static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, Size_t vecpos, goopax_future<gather_type>&, P... p)
	{
	  pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos, p...);
	}
	template <class RP, class... P> static inline void setcollect(rundata* rd, RP rp, goopax_future<gather_type>& b, P... p)
	{
	  pas<R, REST>::setcollect(rd, rp, p...);
	  b.set(rd, pas<R, REST>::gather_size(), collectfunc<G>);
	}
      };


      template <class T, class R, class... A> struct pas<R, tuple<T, A...> >
      {
	using raw_T = typename remove_const<typename remove_reference<T>::type>::type;
	using split_t = pas_split<R, T, tuple<A...>, is_gpu_type<raw_T>::value, is_resource_type<raw_T>::value, is_base_of<gather_basecheck, raw_T>::value>;
	static GOOPAX_CONSTEXPR unsigned int param_size() { return split_t::param_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_size() { return split_t::gather_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_num() { return split_t::gather_num(); }
	static const unsigned int num_tex_args = split_t::num_tex_args;
	template <class... P> static void impl(P&... p)
	{
	  split_t::impl(p...);
	}
	template <class K, class V, typename PARAM, class E, class RP, class... P> static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, Size_t vecpos, E e, P... p)
	{
	  split_t::prepare(device_p, kernel, param, rp, vec, vecpos, *e, p...);
	}
	template <class E, class RP, class... P> static inline void setcollect(rundata* rd, RP rp, E e, P... p)
	{
	  split_t::setcollect(rd, rp, *e, p...);
	}
      };

      
      template <> struct pas<void, tuple<> >
      {
	using return_type = void;
	static GOOPAX_CONSTEXPR unsigned int param_size() { return 0; }
	static GOOPAX_CONSTEXPR unsigned int gather_size() { return 0; }
	static GOOPAX_CONSTEXPR unsigned int gather_num() { return 0; }
	static GOOPAX_CONSTEXPR unsigned int num_tex_args = 0;

	template <class K, class... P> static void impl(K& kernel, P&... p)
	{
	  kernel.program(p...);
	}
	template <class K, class V, typename PARAM> static inline void prepare(goopax_device_impl*, K&, PARAM&, void*, V&, Size_t)
	{
	}
	static inline void setcollect(rundata*, void*)
	{
	}
      };


      template <class R> struct pas<R, tuple<> >
      {
	using return_type = typename R::value_type;
	static GOOPAX_CONSTEXPR unsigned int param_size() { return 0;}
	static GOOPAX_CONSTEXPR unsigned int gather_size() { return pas_data::paramsize<return_type>();}
	static GOOPAX_CONSTEXPR unsigned int gather_num() { return 1;}
	static GOOPAX_CONSTEXPR unsigned int num_tex_args = 0;

	template <class K, class... P> static void impl(K& kernel, P&... p)
	{
	  auto ret = kernel.program(p...);
	  typename gettype<return_type>::gpu tot = work_group_reduce<R>((typename gettype<return_type>::gpu)ret, local_size());
	  gpu_if(local_id() == 0)
	    pas_data::write_gather(group_id()*pas_data::paramsize<return_type>(), tot);
	}
	template <class K, class V, typename PARAM, class RP> static inline void prepare(goopax_device_impl*, K&, PARAM&, RP, V&, Size_t)
	{
	}
	template <class RP> static inline void setcollect(rundata* rd, RP rp)
	{
	  rp->set(rd, 0, collectfunc<R>);
	}
      };


	  template <class G> struct return_type_wrapper
	  {
		  using type = typename G::value_type;
	  };
	  template <> struct return_type_wrapper<void>
	  {
		  struct void_return_type {};

		  using type = void_return_type;
	  };


      template <class FUNC> struct funcinfo;

      template <class K, class R, class... A> struct funcinfo<R (K::*)(A...)>
      {
	  using params = tuple<A...>;
	  using return_type = R;
	static GOOPAX_CONSTEXPR unsigned int param_size() { return pas<R, tuple<A...> >::param_size(); }
	static GOOPAX_CONSTEXPR unsigned int gather_size() { return pas<R, tuple<A...> >::gather_size(); }

	static void impl(K& kernel)
	{
	  pas<R, tuple<A...> >::impl(kernel);
	}
      };


	template <typename T, bool is_gather, bool is_struct> struct change_type_single;

	template <typename T, typename SIZE>
	struct change_type_single<resource<T, SIZE, false>, false, false>
	{
	    using type = buffer<T, SIZE>;
	    static constexpr bool can_get_device = true;
	};

	template <unsigned int DIM, typename T, bool is_normalized>
	struct change_type_single<image_resource<DIM, T, is_normalized>, false, false>
	{
	    using type = image_buffer<DIM, T, is_normalized>;
	    static constexpr bool can_get_device = true;
	};
	
	template <unsigned int DIM, typename T, bool is_normalized>
	struct change_type_single<image_array_resource<DIM, T, is_normalized>, false, false>
	{
	    using type = image_array_buffer<DIM, T, is_normalized>;
	    static constexpr bool can_get_device = true;
	};

	template <typename T>
	struct change_type_single<T&, false, false>
	{
	    using type = typename change_type_single<T, false, false>::type&;
	    static constexpr bool can_get_device = change_type_single<T, false, false>::can_get_device;
	};
	template <typename T>
	struct change_type_single<const T&, false, false>
	{
	    using type = const typename change_type_single<T, false, false>::type&;
	    static constexpr bool can_get_device = change_type_single<T, false, false>::can_get_device;
	};
	
	template <typename T>
	struct change_type_single<T&, true, false>
	{
	    using type = goopax_future<typename T::value_type>&;
	    static constexpr bool can_get_device = false;
	};
	template <typename T>
	struct change_type_single<gpu_type<T>, false, false>
	{
	    using type = const T&;
	    static constexpr bool can_get_device = false;
	};
	
	template <typename T>
	struct change_type_single<T, false, true>
	{
	    using type = const typename unrangetype<typename gettype<T>::cpu>::type&;
	    static constexpr bool can_get_device = false;
	};

	
	
	template <typename...> struct change_type;

	template <typename... HAVE, typename T, typename... B> struct change_type<tuple<HAVE...>, tuple<T, B...>>
	{
	    using raw_T = typename remove_const<typename remove_reference<T>::type>::type;
	    using current = change_type_single<T, is_base_of<gather_basecheck, raw_T>::value, is_gpu_struct<raw_T>::value>;
	    using rest = change_type<tuple<HAVE..., typename current::type>, tuple<B...>>;
	    static constexpr bool can_get_device_from_current = current::can_get_device;
	    using type = typename rest::type;
	    static constexpr bool can_get_device = can_get_device_from_current || rest::can_get_device;
	    
	  template <typename... C> static inline goopax_device_impl* get_device_split(true_type, typename current::type t, C&&...)
	    {
	      return t.get_device().get_impl();
	    }
	  template <typename... C> static inline goopax_device_impl* get_device_split(false_type, typename current::type, C&&... b)
	    {
		return rest::get_device(b...);
	    }
	  template <typename... C> static inline goopax_device_impl* get_device(typename current::type t, C&&... b)
	    {
	      return get_device_split(integral_constant<bool, can_get_device_from_current>(), t, b...);
	    }
	};
	
	template <typename HAVE> struct change_type<HAVE, tuple<>>
	{
	    using type = HAVE;
	    static constexpr bool can_get_device = false;
	    static inline goopax_device_impl* get_device()
	    {
		abort();
	    }
	};

	
	template <typename FUNC, typename P> class kernelparams;
	template <typename K, typename R, typename... A, typename... B> class kernelparams<R (K::*)(A...), tuple<B...>>
	{
	public:
	    using return_type = goopax_future<typename return_type_wrapper<R>::type>;

	    static inline void arun(K& kernel, return_type* ret, B&&... b)
	    {
	      using ct = change_type<tuple<>, tuple<A...>>;
	      goopax_device_impl* device = ct::get_device(b...);
	      return arun(kernel, device, ret, b...);
	    }
	    static inline void arun(K& kernel, goopax_device_impl* device, return_type* ret, B&&... p)
	    {
	      static constexpr unsigned int param_size = pas<R, tuple<A...> >::param_size();
	      //static constexpr unsigned int gather_size = pas<R, tuple<A...> >::gather_size();
	      static constexpr unsigned int gather_num = pas<R, tuple<A...> >::gather_num();
	      static constexpr unsigned int num_tex_args = pas<R, tuple<A...> >::num_tex_args;
	      
	      array<paramtype_t, param_size> pd;
	      array<texbuf*, num_tex_args> vec;
	      pas<R, tuple<A...> >::prepare(device, kernel, pd, ret, vec, 0, &p...);

	      rundata* rd = if_call(interface::goopax_E_fk_arun3, device, kernel.impl_p.get(), vec.data(), (goopax::impl::Uint)vec.size(), pd.data(), gather_num);

	      pas<R, tuple<A...> >::setcollect(rd, ret, &p...);
	      

	    }

	};
	

        /// \ingroup kernel
        /// \brief kernel class to derive gpu kernels from
        /// \param K derived kernel class.
        /// If The C++ language standard is < 14, any return type of the kernel must be specified with the RETURN_TYPE template argument.
        /// All gpu kernels must be derived from this class, where the new class must be passed as template argument, like so:
        /// struct my_kernel :
        /// public kernel<my_kernel>
        /// {
        ///    void program(...) {...}
        /// };
#if GOOPAX_USE_CXX14
      template <class K> struct kernel :
#else
      template <class K, class RETURN_TYPE=void> struct kernel :
#endif
	public full_kernel
      {
	virtual void implcall(goopax::impl::Tuint* p) final override
	{
	  using functype = decltype(&K::program);
	  *p = funcinfo<functype>::param_size();
	  funcinfo<functype>::impl(static_cast<K&>(*this));
	}


 	/// Operator (). This starts the gpu kernel on the device. The parameters must match the arguments specified in the function 'program()', exept that cpu types are passed, where gpu types are specified in the function 'program()'. The device may be specified as a first argument. This is optional except if the device cannot be deduced from the other arguments. All buffers must reside on the same device, which is the device on which the kernel is executed.
#if GOOPAX_USE_CXX14
	template <class... A> inline auto operator () (goopax_device device, A&&... a)
#else
	  template <class... A> inline goopax_future<RETURN_TYPE> operator () (goopax_device device, A&&... a)
#endif
	{
	  using functype = decltype(&K::program);
	    using ct = change_type<tuple<>, typename funcinfo<functype>::params>;
	    using kp = kernelparams<decltype(&K::program), typename ct::type>;
	  goopax_future<typename return_type_wrapper<typename funcinfo<functype>::return_type>::type> ret;
#if !GOOPAX_USE_CXX14
	  static_assert(is_same<RETURN_TYPE, typename funcinfo<functype>::return_type::value_type>::value, "");
#endif
	  kp::arun(static_cast<K&>(*this), device.get_impl(), &ret, a...);
	  return ret;
	}
          /// Operator (). This starts the gpu kernel on the device. The parameters must match the arguments specified in the function 'program()', exept that cpu types are passed, where gpu types are specified in the function 'program()'.
          /// The device is deduced from the buffers that are passed as arguments.
#if GOOPAX_USE_CXX14
	template <class... A> inline auto operator () (A&&... a)
#else
	  template <class... A> inline goopax_future<RETURN_TYPE> operator () (A&&... a)
#endif
	{
	  using functype = decltype(&K::program);
	    using ct = change_type<tuple<>, typename funcinfo<functype>::params>;
	    using kp = kernelparams<decltype(&K::program), typename ct::type>;
	    static_assert(ct::can_get_device, "Cannot deduce device from input buffers. Please pass the goopax_device as first argument.");

	    goopax_future<typename return_type_wrapper<typename funcinfo<functype>::return_type>::type> ret;
	    kp::arun(static_cast<K&>(*this), &ret, a...);

#if !GOOPAX_USE_CXX14
	  static_assert(is_same<RETURN_TYPE, typename funcinfo<functype>::return_type::value_type>::value, "");
#endif
	  return ret;
	}
	  
	kernel()
	{
            /// constructor
	  this->impl_p.reset(if_call(interface::goopax_E_fk_new2,
				     static_cast<full_kernel*>(this),
                                     GOOPAX_CERT, sig,
                                     fk_call_impl<void>,
				     fk_get_name<void>,
				     funcinfo<decltype(&K::program)>::param_size(),
				     funcinfo<decltype(&K::program)>::gather_size()));
	}
	kernel(const kernel&) = delete;
	kernel& operator = (const kernel&) = delete;
          /// move semantics
	kernel(kernel&&) = default;
	kernel& operator = (kernel&& b)
	{
            /// move semantics
	  if (this->impl_p.get() != nullptr)
	    {
	      wait(nullptr);
	    }
	  static_cast<full_kernel&>(*this) = move(b);
	  return *this;
	}
	~kernel()
	{
	  if (this->impl_p.get() != nullptr)
	    {
	      wait(nullptr);
	    }
	}
      };

#if !GOOPAX_USE_CXX14
      template <class K> struct kernel<K, void> :
	public full_kernel
      {
	virtual void implcall(goopax::impl::Tuint* p) final override
	{
	  using functype = decltype(&K::program);
	  *p = funcinfo<functype>::param_size();
	  funcinfo<functype>::impl(static_cast<K&>(*this));
	}

	template <class... A> inline void operator () (goopax_device device, A&&... a)
	{
	  using functype = decltype(&K::program);

	  using ct = change_type<tuple<>, typename funcinfo<functype>::params>;
	  using kp = kernelparams<decltype(&K::program), typename ct::type>;
	  goopax_future<typename return_type_wrapper<void>::type> ret;
	  kp::arun(static_cast<K&>(*this), device.get_impl(), &ret, a...);
	}
	template <class... A> inline void operator () (A&&... a)
	{
	  using functype = decltype(&K::program);
	    using ct = change_type<tuple<>, typename funcinfo<functype>::params>;
	    using kp = kernelparams<decltype(&K::program), typename ct::type>;
	    goopax_future<typename return_type_wrapper<void>::type> ret;
	    static_assert(ct::can_get_device, "Cannot deduce device from input buffers. Please pass the goopax_device as first argument.");
	    kp::arun(static_cast<K&>(*this), &ret, a...);
	}

	kernel()
	{
	  this->impl_p.reset(if_call(interface::goopax_E_fk_new2,
				     static_cast<full_kernel*>(this),
                                     GOOPAX_CERT, sig,
				     fk_call_impl<void>,
				     fk_get_name<void>,
				     funcinfo<decltype(&K::program)>::param_size(),
				     funcinfo<decltype(&K::program)>::gather_size()));
	}

	~kernel()
	{
	  if (this->impl_p.get() != nullptr)
	    {
	      wait(nullptr);
	    }
	}
      };
#endif
    }

    template <typename FUNC, typename OP> struct auto_kernel;

    template <typename FUNC, typename ClassType, typename ReturnType, typename... Args>
    struct auto_kernel<FUNC, ReturnType(ClassType::*)(Args...) const> :
      public ns_param_kernel::kernel<auto_kernel<FUNC, ReturnType(ClassType::*)(Args...) const>>
    {
      FUNC func;
    
      ReturnType program(Args... a)
      {
	return func(a...);
      }
      auto_kernel(FUNC f) :
	func(f)
	{}
    };

    template <typename FUNC, typename ClassType, typename... Args>
    struct auto_kernel<FUNC, void(ClassType::*)(Args...) const> :
      public ns_param_kernel::kernel<auto_kernel<FUNC, void(ClassType::*)(Args...) const>>
    {
      FUNC func;
    
      void program(Args... a)
      {
	func(a...);
      }
      auto_kernel(FUNC f) :
	func(f)
	{}
    };

    template <class FUNC> auto_kernel<FUNC, decltype(&FUNC::operator())> make_kernel(FUNC func)
    {
      using OP = decltype(&FUNC::operator());
      
      return auto_kernel<FUNC, OP>(func);
    }
  }
  
  using impl::ns_param_kernel::kernel;
  using impl::full_kernel;
  using impl::gather_base;
  using impl::make_kernel;
  using impl::goopax_future;
}


/// \}

