// -*- C++ -*-

#pragma once

#include "gpu_types.h"
#include "lib/param.h"


/// \addtogroup reinterpret
/// The reinterpret functions provide functionalities for bit-wise conversion between data types without changing the underlying representation.
/// Various types can be converted into each other: intrinsic types to goopax structs, goopax structs to std::vectors of other goopax structs, vectors to intrinsic types, etc.
/// The sizes of the source value and the destination value must match, otherwise the compiler will report an error. If the source type is an std::vector, the size is checked at runtime.
/// If the destination type is an std::vector, its size is chosen so that it matches the source type.
/// If the source type is an CPU type, the destination type must also be a CPU type. If the source type is a GPU type, the destination type must also be a GPU type.
/// \{

namespace goopax
{
  namespace impl
  {
    namespace reinterpretor
    {
      /**
       \brief bit-wise conversion
       \ingroup reinterpret
	Reinterpret bits into another type.
	\param TO destination type
	\param from source
	This function reinterprets the input parameter into another type.
	FROM and TO must have the same size. They may be fixed-size arrays.
	
	Example: 
	gpu_double a = 17;
	array<gpu_int, 2> b = reinterpret<array<gpu_int, 2> >(a);
	gpu_uint64 c = reinterpret<gpu_uint64> c;
       */ 
      template <class TO, class FROM>
      inline TO reinterpret(const FROM& from);
    }
    
    namespace interface
    {
      extern "C"
      {
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_reinterpret_gpuvec2(gpu_writer_base_impl** to, Size_t Nto, gpu_writer_base_impl** from, Size_t Nfrom, Uint Efrom) noexcept;
	  GOOPAX_DLL_EXPORT(3.1.1) void goopax_reinterpret_struct(const gpu_struct::mapframe_base&, const gpu_struct::mapframe_base&, const gpu_struct::mapframe_base&, const gpu_struct::mapframe_base&, const void*, void*, Bool is_gpu) noexcept;

      }
    }


    namespace reinterpretor
    {
      template <class TO, class FROM, Size_t NTO, Size_t NFROM>
      void reinterpret_switch3(array<gpu_type<TO>, NTO>& to, const array<gpu_type<FROM>, NFROM>& from)
      {
	static_assert(gettype<TO>::size*NTO == gettype<FROM>::size*NFROM, "Reinterpret: Objects have different size.");
	vector<gpu_writer_base_impl*> toi(NTO);
	for (Size_t k=0; k<NTO; ++k)
	  {
	    toi[k] = to[k].impl.get();
	  }
	vector<gpu_writer_base_impl*> fromi(NFROM);
	for (Size_t k=0; k<NFROM; ++k)
	  {
	    fromi[k] = from[k].impl.get();
	  }
	check_errno(interface::goopax_E_reinterpret_gpuvec2(&toi[0], NTO, &fromi[0], NFROM, gettype<FROM>::size));
      }

      
      template <class F, class FA, Size_t NF, class T, class TA, Size_t NT>
      void reinterpret_switch3(array<checknan<T, TA>, NT>& to, const array<checknan<F, FA>, NF>& from, typename std::enable_if<NF==1 || NT==1, int>::type=0)
      {
	Size_t num_uninit = 0;
	for (Size_t k=0; k<NF; ++k)
	  {
	    if (from[k].is_uninitialized())
	      ++num_uninit;
	  }
	if (num_uninit == 0)
	  {
	    convert_assign(to, reinterpret<array<T, NT>>(unrange(from)));
	  }
	else if (num_uninit == NF)
	  to = array<checknan<T, TA>, NT>();
	else
	  {
	    warning() << "GOOPAX: reinterpret: " << num_uninit << " of " << NF << " elements are uninitialized!" << endl;
	    abort();
	  }
      }
      
      template <class F, Size_t NF, class T, class TA, Size_t NT>
      void reinterpret_switch3(array<checknan<T, TA>, NT>& to, const array<F, NF>& from, typename enable_if<(NF==1 || NT==1) && (is_fundamental<F>::value || is_pointer<F>::value), int>::type=0)
      {
	array<T, NT> tmp = reinterpret<array<T, NT>>(from);
	std::copy(tmp.begin(), tmp.end(), to.begin());
      }

      template <class F, class FA, Size_t NF, class T, Size_t NT>
      void reinterpret_switch3(array<T, NT>& to, const array<checknan<F, FA>, NF>& from, typename enable_if<(NF==1 || NT==1) && (is_fundamental<T>::value || is_pointer<T>::value), int>::type=0)
      {
	array<F, NF> tmp;
	std::copy(from.begin(), from.end(), tmp.begin());
	to = reinterpret<array<T, NT>>(tmp);
      }
      

      template <class TO, class FROM>
      inline void reinterpret_switch3(TO& to, const FROM& from)
      {
	static_assert(gettype<TO>::size == gettype<FROM>::size, "Reinterpret: Objects have different size.");
	static_assert(is_same<typename unrangetype<typename get_X_type<FROM>::type>::type, typename unrangetype<typename get_X_type<TO>::type>::type>::value, "");
	const auto& mf_from = gpu_struct::mapframe<FROM>::get_static_mf();
	const auto& mf_fromu = gpu_struct::mapframe<typename goopax_struct_changetype<FROM, Uint>::type>::get_static_mf();
	const auto& mf_to = gpu_struct::mapframe<TO>::get_static_mf();
	const auto& mf_tou = gpu_struct::mapframe<typename goopax_struct_changetype<TO, Uint>::type>::get_static_mf();
	  
	interface::goopax_reinterpret_struct(mf_from, mf_fromu, mf_to, mf_tou, &from, &to, is_same<TO, typename gettype<TO>::gpu>::value);
      }



      template <class TO, class FROM>
      inline void reinterpret_switch3(std::vector<TO>& to, const std::vector<FROM>& from)
      {
	static_assert((gettype<TO>::size%gettype<FROM>::size== 0) ||
		      (gettype<FROM>::size%gettype<TO>::size== 0), "Reinterpret: Object sizes must be divisible.");
		       
	to.resize(from.size() * gettype<FROM>::size / gettype<TO>::size);
	assert(from.size() * gettype<FROM>::size == to.size() * gettype<TO>::size); // FIXME: Allow mismatch, if possible. Allow zero size.

	GOOPAX_CONSTEXPR Size_t bsto = gettype<FROM>::size / gettype<TO>::size == 0 ? 1 : gettype<FROM>::size / gettype<TO>::size;
	GOOPAX_CONSTEXPR Size_t bsfrom = gettype<TO>::size / gettype<FROM>::size == 0 ? 1 : gettype<TO>::size / gettype<FROM>::size;

	auto p_to = to.begin();
	auto p_from = from.begin();
	while (p_to != to.end())
	  {
	    array<FROM, bsfrom> tmp;
	    std::copy(p_from, p_from+bsfrom, tmp.begin());
	    array<TO, bsto> bto = reinterpret<array<TO, bsto> >(tmp);
	    for (Size_t k=0; k<bsto; ++k)
	      {
		p_to[k] = bto[k];
	      }
	    p_to += bsto;
	    p_from += bsfrom;
	  }
#if GOOPAX_DEBUGHARD >= 2
	assert(p_to == to.end());
	assert(p_from == from.end());
#endif
      }
      
      template <typename TO, typename FROM>
      inline void reinterpret_switch3(TO& to, const std::vector<FROM>& from)
      {
	static_assert(gettype<TO>::size % gettype<FROM>::size == 0, "Reinterpret: Object sizes must be divisible.");
	array<FROM, (gettype<TO>::size / gettype<FROM>::size)> tmp;
	std::copy(from.begin(), from.end(), tmp.begin());
	to = reinterpret<TO>(tmp);
      }
      
      template <typename TO, typename FROM>
      inline void reinterpret_switch3(std::vector<TO>& to, const FROM& from)
      {
	static_assert(gettype<FROM>::size % gettype<TO>::size == 0, "Reinterpret: Object sizes must be divisible.");
	auto tmp = reinterpret<array<TO, (gettype<FROM>::size / gettype<TO>::size)>>(from);
	to.assign(tmp.begin(), tmp.end());
      }
      

      template <typename TO, typename FROM>
      inline void reinterpret_switch2(TO& to, const FROM& from)
      {
	reinterpret_switch3(to, array<FROM, 1>({{from}}));
      }

      template <typename TO, typename FROM>
      inline void reinterpret_switch2(TO& to, const std::vector<FROM>& from)
      {
	reinterpret_switch3(to, from);
      }
#if GOOPAX_HAVE_DEBUG_STL
      template <typename TO, typename FROM>
      inline void reinterpret_switch2(TO& to, const std::__debug::vector<FROM>& from)
      {
	reinterpret_switch3(to, std::vector<FROM>(from.begin(), from.end()));
      }
#endif
      
      template <typename TO, typename FROM, Size_t NFROM>
      inline void reinterpret_switch2(TO& to, const goopax::impl::array<FROM, NFROM>& from)
      {
	reinterpret_switch3(to, from);
      }
							       
      
      
      template <typename TO, typename FROM>
      inline void reinterpret_switch(TO& to, const FROM& from, false_type)
      {
	array<TO, 1> tmp;
	reinterpret_switch2(tmp, from);
	to = tmp[0];
      }
				     
      template <typename TO, typename FROM>
      inline void reinterpret_switch(std::vector<TO>& to, const FROM& from, false_type)
      {
	reinterpret_switch2(to, from);
      }
				     
#if GOOPAX_HAVE_DEBUG_STL
      template <typename TO, typename FROM>
      inline void reinterpret_switch(std::__debug::vector<TO>& to, const FROM& from, false_type)
      {
	std::vector<TO> tmp;
	reinterpret_switch2(tmp, from);
	to.assign(tmp.begin(), tmp.end());
      }
#endif      

      template <typename TO, typename FROM, Size_t NTO>
      inline void reinterpret_switch(goopax::impl::array<TO, NTO>& to, const FROM& from, false_type)
      {
	reinterpret_switch2(to, from);
      }
				     
      
      template <class TO, class FROM> 
      inline void reinterpret_switch(TO& to, const FROM& from, true_type)
      {
	static_assert(is_same<TO, typename unrangetype<TO>::type>::value, "");
	static_assert(is_same<FROM, typename unrangetype<FROM>::type>::value, "");

	using TOraw = typename unrangetype<TO>::type;
	using FROMraw = typename unrangetype<FROM>::type;
	static_assert(sizeof(FROMraw) == sizeof(TOraw), "Reinterpret: Objects have different size.");
	const FROMraw fromraw = from;
	const FROMraw* GOOPAX_GCC_ATTRIB(__may_alias__) fromraw_a = &fromraw;
	const TOraw* GOOPAX_GCC_ATTRIB(__may_alias__) toraw_a = reinterpret_cast<const TOraw*>((const void*)fromraw_a);
	to = *toraw_a;
      }




      template <class TO, class FROM>
      inline TO reinterpret(const FROM& from)
      {
	TO to;
	reinterpret_switch(to, from, integral_constant<Bool, (isplain<TO>::value && isplain<FROM>::value)>());
	return to;
      }
    }
    using reinterpretor::reinterpret;

  }
  using impl::reinterpret;
}

/// \}

